@@ -3,7 +3,18 @@ import json
 import copy
 from sklearn.metrics import cohen_kappa_score
 
-def load_data(data_path, require_annotations=True):
+
+def count_annotations_project(project):
+    cnt = 0
+    for doc in project['documents']:
+        for ann in doc['annotations']:
+            # Only validated
+            if ann['validated']:
+                cnt += 1
+    return cnt
+
+
+def load_data(data_path, require_annotations=True, order_by_num_ann=True):
     data_candidates = json.load(open(data_path))
     if require_annotations:
         data = {'projects': []}
@@ -19,9 +30,30 @@ def load_data(data_path, require_annotations=True):
     else:
         data = data_candidates
 
+
+    cnts = []
+    if order_by_num_ann:
+        for project in data['projects']:
+            cnt = count_annotations_project(project)
+            cnts.append(cnt)
+    srt = np.argsort(-np.array(cnts))
+    data['projects'] = [data['projects'][i] for i in srt]
+
     return data
 
 
+def count_annotations(data_path):
+    data = load_data(data_path, require_annotations=True)
+
+    g_cnt = 0
+    for project in data['projects']:
+        cnt = count_annotations_project(project)
+        g_cnt += cnt
+        print("Number of annotations in '{}' is: {}".format(project['name'], cnt))
+
+    print("Total number of annotations is: {}".format(g_cnt))
+
+
 def get_doc_from_project(project, doc_id):
     for document in project['documents']:
         if document['id'] == doc_id:
@@ -40,7 +72,7 @@ def meta_ann_from_ann(ann, meta_name):
     meta_anns = ann['meta_anns']
     # need for old versions of data
     if type(meta_anns) == dict:
-        return meta_anns.get(meta_name, Name)
+        return meta_anns.get(meta_name, None)
     else:
         for meta_ann in meta_anns:
             if meta_ann['name'] == meta_name:
@@ -72,8 +104,10 @@ def are_anns_same(ann, ann2, meta_names=[], require_double_inner=True):
     return True
 
 
-def are_docs_same(document, document2, require_double_inner=True, ann_stats=[], meta_names=[]):
+def get_same_anns(document, document2, require_double_inner=True, ann_stats=[], meta_names=[]):
     are_same = True
+    new_document = copy.deepcopy(document)
+    new_document['annotations'] = []
 
     for ann in document['annotations']:
         # Take only validated anns
@@ -82,13 +116,14 @@ def are_docs_same(document, document2, require_double_inner=True, ann_stats=[],
 
             if ann2 is not None:
                 if not are_anns_same(ann, ann2, meta_names):
-                    print("SFFF")
                     ann_stats.append((1, 0))
                     are_same = False
                 else:
                     ann_stats.append((1, 1))
+                    new_document['annotations'].append(ann)
             elif not require_double_inner:
                 ann_stats.append((1, 1))
+                new_document['annotations'].append(ann)
             else:
                 ann_stats.append((1, 0))
                 are_same = False
@@ -101,36 +136,41 @@ def are_docs_same(document, document2, require_double_inner=True, ann_stats=[],
             if not require_double_inner:
                 ann_stats.append((1, 1))
                 # Also append this annotation to the document, because it is missing from it
-                document['annotations'].append(ann2)
+                new_document['annotations'].append(ann2)
             else:
                 ann_stats.append((0, 1))
                 are_same = False
 
-    return are_same
+    return new_document
 
 
-def count_annotations(data_path):
-    data = load_data(data_path, require_annotations=True)
 
-    g_cnt = 0
-    for project in data['projects']:
-        cnt = 0
-        for doc in project['documents']:
-            for ann in doc['annotations']:
-                # Only validated
-                if ann['validated']:
-                    cnt += 1
-                    g_cnt += 1
-        print("Number of documents in '{}' is: {}".format(project['name'], cnt))
+def print_consolid_stats(ann_stats=[]):
+    if ann_stats:
+        ann_stats = np.array(ann_stats)
+        ck = 0 #TODO: cohen_kappa_score(ann_stats[:, 0], ann_stats[:, 1])
+        t = 0
+        for i in range(len(ann_stats)):
+            if ann_stats[i, 0] == ann_stats[i, 1]:
+                t += 1
+        agr = t / len(ann_stats)
+        print("Kappa: {:.4f}; Agreement: {:.4f}".format(ck, agr)) 
+        print("InAgreement vs Total: {} / {}".format(t, len(ann_stats)))
 
-    print("Total number of documents is: {}".format(g_cnt))
 
-def consolidate_double_annotations(data_path, out_path, require_double=True, require_double_inner=True, meta_anns_to_match=[]):
-    """
+def consolidate_double_annotations(data_path, out_path, require_double=True, require_double_inner=False, meta_anns_to_match=[]):
+    """ Consolidated a dataset that was multi-annotated (same documents two times).
 
+    data_path:
+        Output from MedCATtrainer - projects containig the same documents must have the same name.
+    out_path:
+        The consolidated data will be saved here - usually only annotations where both annotators agre
+            out_path:
+            The consolidated data will be saved here - usually only annotations where both annotators agreee
     require_double (boolean):
         If True everything must be double annotated, meaning there have to be two projects of the same name for each name. Else, it will
-            also use projects thta do not have double annotiations.
+            also use projects that do not have double annotiations. If this is False, projects that do not have double anns will be
+            included as is, and projects that have will still be checked.
     require_double_inner (boolean):
         If False - this will allow some entities to be annotated by only one annotator and not the other, while still requiring
             annotations to be the same if they exist.
@@ -146,6 +186,8 @@ def consolidate_double_annotations(data_path, out_path, require_double=True, req
     for project in data['projects']:
         id = project['id']
         new_documents = []
+        ann_stats_project = []
+        new_project = None
         if id not in projects_done:
             projects_done.add(id)
             name = project['name']
@@ -170,26 +212,31 @@ def consolidate_double_annotations(data_path, out_path, require_double=True, req
                     for document in documents:
                         document2 = get_doc_from_project(project2, document['id'])
 
-                        if document2 is not None and are_docs_same(document, document2, require_double_inner=require_double_inner, ann_stats=ann_stats, meta_names=meta_anns_to_match):
+                        if document2 is not None:
+                            new_document = get_same_anns(document, document2, require_double_inner=require_double_inner, ann_stats=ann_stats_project, meta_names=meta_anns_to_match)
+                            new_documents.append(new_document)
+                        elif not require_double_inner:
+                            # Use the base document if we are allowing no double anns
                             new_documents.append(document)
 
-                new_project = copy.deepcopy(project)
-                new_project['documents'] = new_documents
-            out_data['projects'].append(new_project)
+                    new_project = copy.deepcopy(project)
+                    new_project['documents'] = new_documents
+
+            if new_project is not None:
+                ann_stats.extend(ann_stats_project)
+                out_data['projects'].append(new_project)
+
+            if ann_stats_project:
+                print("**Printing stats for project: {}".format(project['name']))
+                print_consolid_stats(ann_stats_project)
+                print("\n\n")
+            else:
+                print("**Project '{}' did not have double annotations\n\n".format(project['name']))
 
     # Save
     json.dump(out_data, open(out_path, 'w'))
-
-    ann_stats = np.array(ann_stats)
-    ck = cohen_kappa_score(ann_stats[:, 0], ann_stats[:, 1])
-    t = 0
-    for i in range(len(ann_stats)):
-        if ann_stats[i, 0] == ann_stats[i, 1]:
-            t += 1
-    agr = t / len(ann_stats)
-    print("Kappa: {:.4f}; Agreement: {:.4f}".format(ck, agr)) 
-    print("Total vs InAgreement: {} / {}".format(t, len(ann_stats)))
-    return ann_stats, out_data
+    print("**Overall stats")
+    print_consolid_stats(ann_stats)
 
 def validate_ner_data(data_path, cdb, cntx_size=70, status_only=False, ignore_if_already_done=False):
     """ Please just ignore this function, I'm afraid to even look at it
@@ -383,13 +430,14 @@ def validate_ner_data(data_path, cdb, cntx_size=70, status_only=False, ignore_if
 
 
 
-def prepare_from_json(data, cntx_left, cntx_right, tokenizer, lowercase=True, cntx_in_chars=False, tui_filter=None):
+def prepare_from_json(data, cntx_left, cntx_right, tokenizer, lowercase=True, cntx_in_chars=False, tui_filter=None, replace_center=None):
     """ Convert the data from a json format into a CSV-like format for training.
 
     data:  json file from MedCAT
     cntx_left:  size of the context
     cntx_right:  size of the context
     tokenizer:  instance of the <Tokenizer> class from huggingface
+    replace_center:  if not None the center word (concept) will be replaced with whatever is set
 
     return:  {'category_name': [('category_value', 'tokens', 'center_token'), ...], ...}
     """
@@ -419,18 +467,32 @@ def prepare_from_json(data, cntx_left, cntx_right, tokenizer, lowercase=True, cn
                                 # Get the index of the center token
                                 ind = 0
                                 for ind, pair in enumerate(doc_text.offsets):
-                                    if start >= pair[0] and start <= pair[1]:
+                                    if start >= pair[0] and start < pair[1]:
                                         break
 
                                 _start = max(0, ind - cntx_left)
                                 _end = min(len(doc_text.tokens), ind + 1 + cntx_right)
                                 tkns = doc_text.tokens[_start:_end]
                                 cpos = cntx_left + min(0, ind-cntx_left)
+
+                                if replace_center is not None:
+                                    for p_ind, pair in enumerate(doc_text.offsets):
+                                        if start >= pair[0] and start < pair[1]:
+                                            s_ind = p_ind
+                                        if end > pair[0] and end <= pair[1]:
+                                            e_ind = p_ind
+
+                                    ln = e_ind - s_ind
+                                    tkns[cpos:cpos+ln+1] = [replace_center]
+
                             else:
+                                # TODO: Currently not working properly 
                                 _start = max(0, start - cntx_left)
                                 _end = min(len(text), end + cntx_right)
                                 tkns = tokenizer.encode(text[_start:_end]).tokens
 
+                                raise Exception("Not working properly for now")
+
 
                             # Backward compatibility if meta_anns is a list vs dict in the new approach
                             meta_anns = []
